import keras
from teras._src.api_export import teras_export
from teras._src.typing import ActivationType


@teras_export("teras.layers.TransformerFeedForward")
class TransformerFeedForward(keras.layers.Layer):
    """
    Transformer Feed Forward layer as proposed in the original
    Transformers architecture, in the paper,"Attention is all you need",
    with a slight addition of optional `Dropout` layer.

    Reference(s):
        https://arxiv.org/abs/1706.03762

    Args:
        embedding_dim: int, dimensionality of embeddings being used in
            the model
        hidden_dim: int, hidden dimensionality to use. By default,
            it is four-times of the `embedding_dim`.
        activation: str or callable, activation function to use for the
            inner linear layer. Defaults to "relu",
        dropout: float, dropout rate to use for the dropout layer
            that is applied in between the two linear layer.
            Defaults to 0., because the original transformer
            architecture doesn't employ a `Dropout` layer.
    """
    def __init__(self,
                 embedding_dim: int,
                 hidden_dim: int = None,
                 activation: ActivationType = "relu",
                 dropout: float = 0.,
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.hidden_dim = embedding_dim * 4 if hidden_dim is None else hidden_dim
        self.activation = activation
        self.dropout = dropout

        self.inner = keras.layers.Dense(self.hidden_dim,
                                        activation=self.activation,
                                        name="feedforward_inner")
        self.outer = keras.layers.Dense(self.embedding_dim,
                                        name="feedforward_outer")
        self.dropout_layer = keras.layers.Dropout(
                                        self.dropout,
                                        name="feedforward_dropout")

    def build(self, input_shape):
        self.inner.build(input_shape)
        hidden_shape = input_shape[:-1] + (self.hidden_dim,)
        self.outer.build(hidden_shape)

    def call(self, inputs):
        x = self.inner(inputs)
        x = self.dropout_layer(x)
        return self.outer(x)

    def get_config(self):
        config = super().get_config()
        config.update({
            "embedding_dim": self.embedding_dim,
            "hidden_dim": self.hidden_dim,
            "activation": self.activation,
            "dropout": self.dropout
        })
        return config



# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keras.backend import backend


if backend() == "tensorflow":
    from teras._src.backend.tensorflow import *

elif backend() == "torch":
    from teras._src.backend.torch import *

elif backend() == "jax":
    from teras._src.backend.jax import *

else:
    raise ValueError("Backend not support. "
                     "Expected one of ['tensorflow', 'jax', 'torch']. "
                     f"Found {backend()}")



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.ops.norm")
def norm(x: keras.KerasTensor,
         ord: str = None,
         axis: int = None,
         keepdims: bool = False
         ):
    """
    Matrix or vector norm.

    Args:
        x: Input tensor.
        ord: Order of norm.
        axis: If axis is an integer, it specifies the axis of x along
            which to compute the vector norms. If axis is a 2-tuple,
            it  specifies the axes that hold 2-D matrices, and the matrix
            norms of these matrices are computed. If axis is None then
            either a vector norm (when x is 1-D) or a matrix norm
            (when x is 2-D) is returned. The default is None.
        keepdims: If this is set to True, the axes which are normed over
            are left in the result as dimensions with size one. With this
            option the result will broadcast correctly against the
            original x.

    Returns:
        Norm of the matrix or vector(s).
    """
    return backend.norm(x, ord=ord, axis=axis, keepdims=keepdims)



import keras
from typing import Union, List, Callable, Tuple
import numpy as np
import pandas as pd


# Taken from TensorFlow Addons - Sorry!

Number = Union[
                float,
                int,
                np.float16,
                np.float32,
                np.float64,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ]


ActivationType = Union[str,
                       Callable,
                       keras.layers.Layer,
                ]


IntegerSequence = Union[List[int], Tuple[int]]

LayersList = List[keras.layers.Layer]

LayersCollection = Union[keras.layers.Layer,
                         List[keras.layers.Layer],
                         keras.Model]

NormalizationType = Union[keras.layers.Layer, str]

FloatSequence = Union[List[float],
                      Tuple[float]]

IntegerSequence = Union[List[int],
                        Tuple[int]]

InitializerType = Union[str, keras.initializers.Initializer]


LayerOrModelType = Union[keras.layers.Layer, keras.Model]

FeaturesNamesType = Union[List[str], Tuple[str]]

OptionalIntOrStr = Union[int, str, None]

ListOrArray = Union[list, np.array]

DataFrameOrNdArray = Union[pd.DataFrame, np.ndarray]



import keras
from keras import random, ops
from keras.backend import floatx

from teras._src.layers.layer_list import LayerList
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.GAINGenerator")
class GAINGenerator(keras.Model):
    """
    Generator model for the GAIN architecture proposed by
    Jinsung Yoon et al. in the paper
    "GAIN: Missing Data Imputation using Generative Adversarial Nets."

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        data_dim: int, The dimensionality of the input dataset.
            Note the dimensionality must be equal to the dimensionality of
            the transformed dataset that is passed to the fit method and not
            that of original dataset as the dimensionality of raw input dataset
            may change during transformation.
            One way to access the dimensionality of the transformed dataset is
            through the `.data_dim` attribute of the `GAINDataSampler`
            instance used in sampling the dataset.
        hidden_dims: list, A list of hidden dimensionalities for constructing
            hidden block. For each value, a `Dense` layer of that
            dimensionality is added to the hidden block.
            By default, `units_values` = [`data_dim`, `data_dim`].
        activation_hidden: Activation function to use for the hidden layers
            in the hidden block. Defaults to "relu".
        activation_out: Activation function to use for the output layer of
            the Discriminator. Defaults to "sigmoid"
    """
    def __init__(self,
                 data_dim: int,
                 hidden_dims: IntegerSequence = None,
                 activation_hidden: str = "relu",
                 activation_out: str = "sigmoid",
                 **kwargs):
        super().__init__(**kwargs)

        if hidden_dims is not None and not isinstance(hidden_dims,
                                                      (list, tuple)):
            raise ValueError(
                "`units_values` must be a list or tuple of units which "
                "determines the number of Discriminator blocks and the "
                f"dimensionality of those blocks. But {hidden_dims} was "
                "passed.")

        self.data_dim = data_dim
        self.hidden_dims = hidden_dims
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out

        if self.hidden_dims is None:
            self.hidden_dims = [self.data_dim] * 2
        self.hidden_block = []
        for dim in self.hidden_dims:
            self.hidden_block.append(
                keras.layers.Dense(
                    units=dim,
                    activation=self.activation_hidden,
                    kernel_initializer="glorot_normal",
                )
            )
        self.hidden_block = LayerList(
            self.hidden_block,
            name="generator_hidden_block"
        )
        self.output_layer = keras.layers.Dense(
            self.data_dim,
            activation=self.activation_out,
            name="generator_output_layer")

    def build(self, input_shape):
        self.hidden_block.build(input_shape)
        input_shape = self.hidden_block.compute_output_shape(input_shape)
        self.output_layer.build(input_shape)

    def call(self, inputs, **kwargs):
        # inputs is the concatenation of `mask` and `original data`
        # where `mask` has the same dimensions as data
        # so the inputs received are 2x the dimensions of `original data`
        outputs = self.hidden_block(inputs)
        outputs = self.output_layer(outputs)
        return outputs

    def predict_step(self, data):
        """
        Args:
            Transformed data.
        Returns:
            Raw imputed data that should be reverse transformed to its
            original form using data transformer instance that was used to
            transform data for training.
        """

        data = ops.cast(data, floatx())
        # Create mask
        mask = 1. - ops.cast(ops.isnan(data), dtype=floatx())
        data = ops.where(ops.isnan(data), x1=0., x2=data)
        # Sample noise
        z = random.uniform(ops.shape(data), minval=0., maxval=0.01)
        x = mask * data + (1 - mask) * z
        imputed_data = self(ops.concatenate([x, mask], axis=1))
        imputed_data = mask * data + (1 - mask) * imputed_data
        return imputed_data

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.data_dim,)

    def get_config(self):
        config = super().get_config()
        config.update(
            {"data_dim": self.data_dim,
             "hidden_dims": self.hidden_dims,
             "activation_hidden": self.activation_hidden,
             "activation_out": self.activation_out,
             }
        )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)



import keras
from teras._src.activations import sparsemax
from teras._src.api_export import teras_export


@teras_export("teras.layers.TabNetAttentiveTransformer")
class TabNetAttentiveTransformer(keras.layers.Layer):
    """
    TabNetAttentiveTransformer layer proposed by Arik et al. in the
    "TabNet: Attentive Interpretable Tabular Learning" paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        data_dim: int, dimensionality of the dataset
        batch_momentum: float, batch momentum
    """
    def __init__(self,
                 data_dim: int,
                 batch_momentum: float,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.batch_momentum = batch_momentum

        self.dense = keras.layers.Dense(data_dim,
                                        use_bias=False)
        self.batch_norm = keras.layers.BatchNormalization(
            momentum=self.batch_momentum)

    def build(self, input_shape):
        self.dense.build(input_shape)
        input_shape = self.dense.compute_output_shape(input_shape)
        self.batch_norm.build(input_shape)
        self.built = True

    def call(self, inputs, prior_scales):
        x = self.dense(inputs)
        x = self.batch_norm(x)
        x = x * prior_scales
        x = sparsemax(x)
        return x

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.data_dim,)

    def get_config(self):
        config = super().get_config()
        config.update({
            "data_dim": self.data_dim,
            "batch_momentum": self.batch_momentum
        })
    


import tensorflow as tf
import keras
from teras._src.backend.common.models.pretrainers.tab_transformer import (
    BaseTabTransformerMLMPretrainer,
    BaseTabTransformerRTDPretrainer
)
from keras import ops


class TabTransformerMLMPretrainer(BaseTabTransformerMLMPretrainer):
    """
    MLM version of pretrainer for TabTransformer.

    Args:
        model: keras.Model, instance of `TabTransformerBackbone` model to
            pretrain.
        data_dim: int, number of features in the dataset
        missing_rate: float, percentage of features to make missing.
            Defaults to 0.3 (or 30%)
        mask_seed: int, seed for generating mask. Defaults to 1337
    """
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 missing_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         missing_rate=missing_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, _ = input_shape
        num_features_to_miss = int(self.missing_rate * self.data_dim)
        mask = tf.ones((num_features_to_miss,), dtype=tf.int32)
        mask = tf.pad(mask,
                      [[0, self.data_dim - num_features_to_miss]])
        mask = tf.tile(tf.expand_dims(mask, 0), (batch_size, 1))
        mask = tf.map_fn(lambda x: tf.random.shuffle(x,
                                                     seed=self.mask_seed),
                         mask,
                         dtype=tf.int32)
        return mask

    def train_step(self, data):
        mask = self._create_mask(ops.shape(data))
        mask = ops.cast(mask, data.dtype)
        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        with tf.GradientTape() as tape:
            y_pred = self(data, mask=(1 - mask))
            loss = self.compute_loss(y=data * mask,
                                     y_pred=y_pred)
        # Compute grads
        gradients = tape.gradient(loss, self.trainable_variables)
        # Optimize
        self.optimizer.apply(
            gradients,
            self.trainable_variables
        )

        logs = {m.name: m.result() for m in self.metrics}
        return logs


class TabTransformerRTDPretrainer(BaseTabTransformerRTDPretrainer):
    """
    RTD version of pretrainer for TabTransformer.

    Args:
        model: keras.Model, instance of `TabTransformerBackbone` model to
            pretrain.
        data_dim: int, number of features in the dataset
        replace_rate: float, percentage of features to replace.
            Defaults to 0.3 (or 30%)
        mask_seed: int, seed for generating mask. Defaults to 1337
    """
    def __init__(self,
                 model: keras.Model,
                 data_dim: int,
                 replace_rate: float = 0.3,
                 mask_seed: int = 1337,
                 **kwargs):
        super().__init__(model=model,
                         data_dim=data_dim,
                         replace_rate=replace_rate,
                         mask_seed=mask_seed,
                         **kwargs)

    def _create_mask(self, input_shape):
        if len(input_shape) != 2:
            raise ValueError(
                "Expected `input_shape` to have a rank of 2. "
                f"Received, {input_shape} with rank {len(input_shape)}")
        batch_size, _ = input_shape
        num_features_to_replace = int(self.replace_rate * self.data_dim)
        mask = tf.ones((num_features_to_replace,), dtype=tf.int32)
        mask = tf.pad(mask,
                      [[0, self.data_dim - num_features_to_replace]])
        mask = tf.tile(tf.expand_dims(mask, 0), (batch_size, 1))
        mask = tf.map_fn(lambda x: tf.random.shuffle(x,
                                                     seed=self.mask_seed),
                         mask,
                         dtype=tf.int32)
        return mask

    def train_step(self, data):
        mask = self._create_mask(ops.shape(data))
        mask = ops.cast(mask, data.dtype)
        # In mask, 1 indicates that the feature will be missing, while 0
        # indicates the opposite
        with tf.GradientTape() as tape:
            y_pred = self(data, mask=(1 - mask))
            loss = self.compute_loss(y=mask,
                                     y_pred=y_pred)
        # Compute grads
        gradients = tape.gradient(loss, self.trainable_variables)
        # Optimize
        self.optimizer.apply(
            gradients,
            self.trainable_variables
        )

        logs = {m.name: m.result() for m in self.metrics}
        return logs





from teras._src.layers.activation import GumbelSoftmax
import tensorflow as tf


def test_gumble_softmax_valid_call():
    gumble_softmax = GumbelSoftmax()
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = gumble_softmax(inputs)


def test_gumble_softmax_output_shape():
    gumble_softmax = GumbelSoftmax()
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = gumble_softmax(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 4






import os

import keras
from keras import ops
from keras.src.testing.test_case import TestCase

from teras._src.models.gans.pcgain.classifier import PCGAINClassifier


class PCGAINClassifierTest(TestCase):
    def setUp(self):
        self.inputs = ops.ones((8, 16))

    def test_valid_call(self):
        classifier = PCGAINClassifier(num_classes=5,
                                      data_dim=7)
        outputs = classifier(self.inputs)

    def test_valid_output_shape(self):
        classifier = PCGAINClassifier(num_classes=5,
                                      data_dim=7)
        outputs = classifier(self.inputs)
        assert ops.shape(outputs) == (8, 5)

    def test_save_and_load(self):
        classifier = PCGAINClassifier(num_classes=5,
                                      data_dim=7)
        outputs_original = classifier(self.inputs)
        save_path = os.path.join(self.get_temp_dir(),
                                 "pcgain_classifier.keras")
        classifier.save(save_path)
        reloaded_model = keras.models.load_model(save_path)
        # Check we got the real object back
        self.assertIsInstance(reloaded_model, PCGAINClassifier)
        outputs_reloaded = reloaded_model(self.inputs)
        self.assertAllClose(
            ops.convert_to_numpy(outputs_original),
            ops.convert_to_numpy(outputs_reloaded)
        )



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.CTGAN")
class CTGAN(backend.models.CTGAN):
    """
    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        generator: keras.Model, An instance of :py:class:`CTGANGenerator`.
        discriminator: keras.Model, An instance of
            :py:class:`CTGANDiscriminator`.
        metadata: dict, A dictionary containing features metadata computed
            during the data transformation step.
            It can be accessed through the `.metadata` property attribute of
            the :py:class:`CTGANDataTransformer` instance which was used to
            transform the raw input data.
            Note that, this is NOT the same metadata as `features_metadata`,
            which is computed using the `get_metadata_for_embedding` utility
            function from :py:mod:`teras.utils`.
        latent_dim: int, Dimensionality of noise or `z` that serves as
            input to :py:class:`CTGANGenerator` to generate samples.
            Defaults to 128.
        seed: int, Seed for random sampling. Defaults to 1337.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 metadata: dict,
                 latent_dim: int = 128,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         metadata=metadata,
                         latent_dim=latent_dim,
                         seed=seed,
                         **kwargs)



from teras._src import backend
from teras._src.typing import IntegerSequence
from teras._src.api_export import teras_export


@teras_export("teras.models.CTGANGenerator")
class CTGANGenerator(backend.models.CTGANGenerator):
    """
    CTGANGenerator for CTGAN architecture as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        data_dim: int, The dimensionality of the dataset.
            It will also be the dimensionality of the output produced
            by the generator.
            Note the dimensionality must be equal to the dimensionality of
            dataset that is passed to the fit method and not necessarily
            the dimensionality of the raw input dataset as sometimes
            data transformation alters the dimensionality of the dataset.
        metadata: dict, `CTGANGenerator` applies different activation functions
            to its outputs depending on the type of features (categorical or
            continuous). And to determine the feature types and for other
            computations during the activation step, the ``metadata``
            computed during the data transformation step, is required.
            It can be accessed through the `.metadata` property attribute of
            the `CTGANDataTransformer` instance which was used to transform
            the raw input data.
            Note that, this is NOT the same metadata as `features_metadata`,
            which is computed using the `get_metadata_for_embedding` utility
            function from `teras.utils`.
            You must access it through the `.metadata` property attribute of the
            `CTGANDataTransformer`.
        hidden_dims: Sequence, A sequence of integers that is used to
            construct the hidden block.
            For each value, a `CTGANGeneratorLayer` of that dimensionality is
            added. Defaults to [256, 256]
    """
    def __init__(self,
                 data_dim: int,
                 metadata: dict,
                 hidden_dims: IntegerSequence = (256, 256),
                 seed: int = 1337,
                 **kwargs):
        super().__init__(data_dim=data_dim,
                         metadata=metadata,
                         hidden_dims=hidden_dims,
                         seed=seed,
                         **kwargs)



from setuptools import setup, find_packages

setup(
    name='teras',
    version='0.3.1',
    description='A Unified Deep Learning Library for Tabular Data.',
    author='Khawaja Abaid',
    author_email='khawaja.abaid@gmail.com',
    packages=find_packages(),
    install_requires=[
        'keras',
        'pandas',
        'scikit-learn',
        'numpy',
    ],
)


from keras import ops
from keras.backend import floatx
from teras._src.api_export import teras_export


@teras_export("teras.losses.tabnet_reconstruction_loss")
def tabnet_reconstruction_loss(real=None,
                               reconstructed=None,
                               mask=None):
    """
    Reconstruction loss for TabNet Pretrainer mode as proposed by
    Sercan et al. in the paper,
    "TabNet: Attentive Interpretable Tabular Learning"

     Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        real: Samples drawn from the input dataset
        reconstructed: Samples reconstructed by the decoder
        mask: Mask that indicates the missing-ness of features in a sample

    Returns:
        Reconstruction loss for TabNet Pretraining.
    """
    nominator_part = (reconstructed - real) * mask
    real_samples_population_std = ops.std(ops.cast(real, dtype=floatx()))
    # divide
    x = nominator_part / real_samples_population_std
    # Calculate L2 norm
    loss = ops.sqrt(ops.sum(ops.square(x)))
    return loss



import keras

from teras._src import backend
from teras._src.api_export import teras_export


@teras_export("teras.models.GAIN")
class GAIN(backend.models.GAIN):
    """
    GAIN is a missing data imputation model based on GANs. This is an
    implementation of the GAIN architecture proposed by Jinsung Yoon et al.
    in the paper,
    "GAIN: Missing Data Imputation using Generative Adversarial Nets"

    In GAIN, the generator observes some components of a real data vector,
    imputes the missing components conditioned on what is actually observed, and
    outputs a completed vector.
    The discriminator then takes a completed vector and attempts to determine
    which components were actually observed and which were imputed. It also
    utilizes a novel hint mechanism, which ensures that generator does in
    fact learn to generate samples according to the true data distribution.

    Reference(s):
        https://arxiv.org/abs/1806.02920

    Args:
        generator: keras.Model, An instance of `GAINGenerator` model or any
            customized model that can work in its place.
        discriminator: keras.Model, An instance of `GAINDiscriminator` model
            or any customized model that can work in its place.
        hint_rate: float, Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Must be between 0. and 1.
            Hint vectors ensure that generated samples follow the underlying
            data distribution.
            Defaults to 0.9
        alpha: float, Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely,
            `generator_loss` = `cross_entropy_loss` + `alpha` * `mse_loss`
            The higher the `alpha`, the more the mse_loss will affect the
            overall generator loss.
            Defaults to 100.
    """
    def __init__(self,
                 generator: keras.Model,
                 discriminator: keras.Model,
                 hint_rate: float = 0.9,
                 alpha: float = 100.,
                 **kwargs):
        super().__init__(generator=generator,
                         discriminator=discriminator,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         **kwargs)



import torch
import keras
from keras import ops, random
from teras._src.backend.common.models.pretrainers.tabnet import BaseTabNetPretrainer


class TabNetPretrainer(BaseTabNetPretrainer):
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 missing_feature_probability: float = 0.3,
                 **kwargs):
        super().__init__(
            encoder=encoder,
            decoder=decoder,
            missing_feature_probability=missing_feature_probability,
            **kwargs)

    def train_step(self, data):
        # Sample mask
        mask = random.binomial(
            shape=ops.shape(data),
            counts=1,
            probabilities=self.missing_feature_probability,
            seed=self._seed_generator)
        # Call torch.nn.Module.zero_grad() to clear the leftover gradients
        # for the weights from the previous train step.
        self.zero_grad()

        # Call
        reconstructed = self(data, mask)

        # Compute loss
        loss = self.compute_loss(
            x=data,
            x_reconstructed=reconstructed,
            mask=mask
        )

        # Compute gradients
        loss.backward()
        trainable_variables = [v for v in self.trainable_variables]
        gradients = [v.value.grad for v in trainable_variables]

        # Update weights
        with torch.no_grad():
            self.optimizer.apply(
                grads=gradients,
                trainable_variables=trainable_variables,
            )

        # Update metrics
        for metric in self.metrics:
            if (metric.name == "loss" or
                    metric.name == "reconstruction_loss"):
                metric.update_state(loss)
            else:
                metric.update_state(data, reconstructed)

        logs = {m.name: m.result() for m in self.metrics}
        return logs



import keras
from keras import ops
from teras._src.api_export import teras_export


@teras_export("teras.layers.CategoricalExtraction")
class CategoricalExtraction(keras.layers.Layer):
    """
    CategoricalExtraction layer extracts categorical features from inputs
    as is. It helps us build functional model where inputs to the
    model contain both categorical and continuous features, but they
    must diverge into two different branches for separate processing.

    Args:
        categorical_idx: list, list of indices of categorical features in the
            given dataset.
    """
    def __init__(self,
                 categorical_idx: list,
                 **kwargs):
        super().__init__(trainable=False, **kwargs)
        self.categorical_idx = categorical_idx

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        output_shape[1] = len(self.categorical_idx)
        return tuple(output_shape)

    def call(self, inputs):
        categorical_features = ops.take(inputs,
                                        indices=self.categorical_idx,
                                        axis=1)
        return categorical_features

    def get_config(self):
        config = super().get_config()
        config.update({
            "categorical_idx": self.categorical_idx
        })



import jax
from keras import ops
from teras._src.backend.common.models.gans.ctgan.generator import BaseCTGANGenerator
from teras._src.typing import IntegerSequence


class CTGANGenerator(BaseCTGANGenerator):
    def __init__(self,
                 data_dim: int,
                 metadata: dict,
                 hidden_dims: IntegerSequence = (256, 256),
                 seed: int = 1337,
                 **kwargs):
        super().__init__(data_dim=data_dim,
                         metadata=metadata,
                         hidden_dims=hidden_dims,
                         seed=seed,
                         **kwargs)

    def apply_activations_by_feature_type(self, interim_outputs):
        """
        This function applies activation functions to the interim outputs
        of the Generator by feature type.
        As CTGAN architecture requires specific transformations on the raw
        input data,
        that decompose one feature in several features,
        and since each type of feature, i.e. continuous or categorical
        require different activation functions to be applied, the process
        of applying those activations becomes rather tricky as it
        requires knowledge of underlying data transformation and
        features metadata.
        To ease the user's burden, in case a user wants to subclass this
        Generator model and completely customize the inner workings of the
        generator but would want to use the activation method specific
        to the CTGAN architecture, so that the subclassed Generator can
        work seamlessly with the rest of the architecture and there
        won't be any discrepancies in outputs produced by the subclasses
        Generator and those expected by the architecture,
        this function is separated, so user can just call this function on
        the interim outputs in the `call` method.

        Args:
            interim_outputs: Outputs produced by the `output_layer` of the
            Generator.

        Returns:
            Final outputs activated by the relevant activation functions.
        """
        outputs = []
        continuous_features_relative_indices = (
            self.metadata)["continuous"]["relative_indices_all"]
        features_relative_indices_all = (
            self.metadata)["relative_indices_all"]
        num_valid_clusters_all = (
            self.metadata)["continuous"]["num_valid_clusters_all"]
        cont_i = 0
        cat_i = 0
        num_categories_all = (
            self.metadata)["categorical"]["num_categories_all"]
        for i, index in enumerate(features_relative_indices_all):
            # the first k = num_continuous_features are continuous in the
            # data
            if i < len(continuous_features_relative_indices):
                # each continuous features has been transformed into
                # num_valid_clusters + 1 features
                # where the first feature is alpha while the following
                # features are beta components
                alphas = ops.tanh(interim_outputs[:, index])
                alphas = ops.expand_dims(alphas, 1)
                outputs.append(alphas)
                start_idx = index + 1
                slice_size = num_valid_clusters_all[cont_i]
                betas = self.gumbel_softmax(
                    jax.lax.dynamic_slice_in_dim(interim_outputs,
                                                 start_index=start_idx,
                                                 slice_size=slice_size,
                                                 axis=1),
                )
                outputs.append(betas)
                cont_i += 1
            # elif index in categorical_features_relative_indices:
            else:
                # each categorical feature has been converted into a
                # one hot vector
                # of size num_categories
                start_idx = index
                slice_size = num_categories_all[cat_i]
                ds = self.gumbel_softmax(
                    jax.lax.dynamic_slice_in_dim(interim_outputs,
                                                 start_index=start_idx,
                                                 slice_size=slice_size,
                                                 axis=1))
                outputs.append(ds)
                cat_i += 1
        outputs = ops.concatenate(outputs, axis=1)
        return outputs



import keras
from keras import random, ops
from teras._src.utils import clean_reloaded_config_data
from teras._src.decorators import assert_fitted


class BaseTVAE(keras.Model):
    """
    Base TVAE class.
    """
    def __init__(self,
                 encoder: keras.Model,
                 decoder: keras.Model,
                 metadata: dict,
                 data_dim: int,
                 latent_dim: int = 128,
                 loss_factor: float = 2.,
                 seed: int = 1337,
                 **kwargs):
        super().__init__(**kwargs)
        self.encoder = encoder
        self.decoder = decoder
        self.metadata = metadata
        self.data_dim = data_dim
        self.latent_dim = latent_dim
        self.loss_factor = loss_factor
        self.seed = seed
        self._seed_gen = random.SeedGenerator(self.seed)
        self.loss_tracker = keras.metrics.Mean(name="loss")

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self._input_shape = input_shape
        if not self.encoder.built:
            self.encoder.build(input_shape)
        input_shape = input_shape[:-1] + (self.latent_dim,)
        if not self.decoder.built:
            self.decoder.build(input_shape)

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    @property
    def metrics(self):
        return [self.loss_tracker]

    def compute_loss(self, real_samples=None, generated_samples=None,
                     sigmas=None, mean=None, log_var=None):
        """
        Evidence Lower Bound (ELBO) Loss [1] adapted for
        TVAE architecture [2] proposed by Lei Xu et al. in the paper,
        "Modeling Tabular data using Conditional GAN".

        Reference(s):
            [1]: https://arxiv.org/abs/1312.6114
            [2]: https://arxiv.org/abs/1907.00503

        Args:
            real_samples: Samples drawn from the input dataset.
            generated_samples: Samples generated by the decoder.
            sigmas: Sigma values returned by decoder
            mean: Mean values returned by encoder
            log_var: Log var values returned by encoder

        Returns:
            Elbo loss adapted for the `TVAE` model.
        """
        loss = []
        cross_entropy = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True,
            reduction="sum")
        cont_i = 0  # continuous index
        cat_i = 0  # categorical index
        for i, relative_index in enumerate(self.metadata["relative_indices_all"]):
            # the first k features are continuous
            if i < self.metadata["num_continuous"]:
                # each continuous feature is of the form
                # [alpha, beta1, beta2...beta(n)] where n is the number of
                # clusters

                # calculate alpha loss
                std = sigmas[relative_index]
                eq = real_samples[:, relative_index] - ops.tanh(
                    generated_samples[:, relative_index])
                loss_temp = ops.sum((eq ** 2 / 2 / (std ** 2)))
                loss.append(loss_temp)
                loss_temp = ops.log(std) * ops.cast(
                    ops.shape(real_samples)[0], dtype=std.dtype)
                loss.append(loss_temp)

                # calculate betas loss
                num_clusters = self.metadata["continuous"][
                    "num_valid_clusters_all"][cont_i]
                logits = generated_samples[:, relative_index + 1: relative_index + 1 + num_clusters]
                labels = ops.argmax(
                    real_samples[:, relative_index + 1: relative_index + 1 + num_clusters],
                    axis=-1)
                cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
                loss.append(cross_entropy_loss)
                cont_i += 1
            else:
                num_categories = self.metadata["categorical"][
                    "num_categories_all"][cat_i]
                logits = generated_samples[:, relative_index: relative_index + num_categories]
                labels = ops.argmax(
                    real_samples[:, relative_index: relative_index + num_categories],
                    axis=-1)
                cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
                loss.append(cross_entropy_loss)
                cat_i += 1
        KLD = -0.5 * ops.sum(1 + log_var - mean ** 2 - ops.exp(log_var))
        loss_1 = ops.sum(loss) * self.loss_factor / self.data_dim
        loss_2 = KLD / self.data_dim
        final_loss = loss_1 + loss_2
        return final_loss

    def call(self, inputs):
        mean, log_var = self.encoder(inputs)
        std = ops.exp(0.5 * log_var)
        eps = random.uniform(shape=ops.shape(std),
                             minval=0, maxval=1,
                             dtype=std.dtype,
                             seed=self._seed_gen)
        z = (std * eps) + mean
        generated_samples, sigmas = self.decoder(z)
        return generated_samples, sigmas, mean, log_var

    def test_step(self, data):
        generated_samples, sigmas, mean, log_var = self(data)
        loss = self.compute_loss(real_samples=data,
                                 generated_samples=generated_samples,
                                 sigmas=sigmas,
                                 mean=mean,
                                 log_var=log_var)
        self.loss_tracker.update_state(loss)
        logs = {m.name: m.result() for m in self.metrics}
        return logs

    def predict_step(self, z):
        generated_samples, _ = self.decoder(z)
        return generated_samples

    def compute_output_shape(self, input_shape):
        batch_size, _ = input_shape
        return ((batch_size, self.data_dim),
                (batch_size, self.latent_dim),
                (self.data_dim,),
                (batch_size, self.latent_dim),
                (batch_size, self.latent_dim))

    def get_config(self):
        config = super().get_config()
        config.update({
            'encoder': keras.layers.serialize(self.encoder),
            'decoder': keras.layers.serialize(self.decoder),
            'metadata': self.metadata,
            'data_dim': self.data_dim,
            'latent_dim': self.latent_dim,
            'loss_factor': self.loss_factor,
            'seed': self.seed,
        })
        return config

    @classmethod
    def from_config(cls, config):
        encoder = keras.layers.deserialize(config.pop("encoder"))
        decoder = keras.layers.deserialize(config.pop("decoder"))
        metadata = clean_reloaded_config_data(config.pop("metadata"))
        return cls(encoder=encoder,
                   decoder=decoder,
                   metadata=metadata,
                   **config)

    def get_build_config(self):
        build_config = dict(input_shape=self._input_shape)
        return build_config

    def build_from_config(self, build_config):
        self.build(build_config["input_shape"])



from keras.src.testing.test_case import TestCase
from keras import ops, random
from teras._src.layers.ctgan.generator_layer import CTGANGeneratorLayer


class CTGANGeneratorLayerTest(TestCase):
    def setUp(self):
        self.input_shape = (4, 8)
        self.inputs = random.normal(shape=self.input_shape)

    def test_valid_call(self):
        encoder_layer = CTGANGeneratorLayer(dim=16)
        outputs = encoder_layer(self.inputs)

    def test_valid_output_shape(self):
        encoder_layer = CTGANGeneratorLayer(dim=16)
        outputs = encoder_layer(self.inputs)
        self.assertEqual(ops.shape(outputs), (4, 8 + 16))



import time
import jax
import numpy as np
import keras
from keras import random, ops
from teras._src.backend.generic_utils import dataset_type
from teras._src.losses.ctgan import ctgan_generator_loss, ctgan_discriminator_loss

# If it works, it works.

# Define variables for CTGAN
_generator: keras.Model = None
_discriminator: keras.Model = None
_metadata = None
_latent_dim = None
_seed = None
_generator_optimizer = None
_discriminator_optimizer = None
_models_built = False
_generator_loss_tracker = keras.metrics.Mean(name="generator_loss")
_discriminator_loss_tracker = keras.metrics.Mean(name="discriminator_loss")

_initialized: bool = False
_compiled: bool = False


def init(generator: keras.Model, discriminator: keras.Model,
         metadata: dict, latent_dim: int = 128, seed: int = 1337):
    """
    `init` method analogous to `keras.Model.__init__()`.


    Args:
        generator: keras.Model, instance of `CTGANGenerator`
        discriminator: keras.Model, instance of `CTGANDiscrimniator`.
        metadata: dict, A dictionary containing features metadata computed
            during the data transformation step.
            It can be accessed through the `.metadata` property attribute of
            the `CTGANDataTransformer` instance which was used to transform
            the raw input data.
            Note that, this is NOT the same metadata as `features_metadata`,
            which is computed using the `get_metadata_for_embedding` utility
            function from `teras.utils`.
        latent_dim: int, Dimensionality of noise or `z` that serves as
            input to `CTGANGenerator` to generate samples. Defaults to 128.
        seed: int, seed for random ops.
    """
    global _generator
    global _discriminator
    global _metadata
    global _latent_dim
    global _seed
    global _initialized
    _generator = generator
    _discriminator = discriminator
    _metadata = metadata
    _latent_dim = latent_dim
    _seed = seed

    _initialized = True


def compile(generator_optimizer=keras.optimizers.Adam(
    name="generator_optimizer"),
        discriminator_optimizer=keras.optimizers.Adam(
            name="discriminator_optimizer")):
    """
    Compile method analogous to `keras.Model.compile()`.

    Args:
        generator_optimizer: Optimizer for generator.
        discriminator_optimizer: Optimizer for discriminator.
    """
    global _generator_optimizer
    global _discriminator_optimizer
    global _compiled
    if not _initialized:
        raise RuntimeError(
            "Trainer not initialized. Please call `Trainer.init` method "
            "before calling `compile`. "
        )

    _generator_optimizer = generator_optimizer
    _discriminator_optimizer = discriminator_optimizer

    _compiled = True


def build(input_shape):
    """
    `build` method analogous to `keras.Model.build()`.

    Args:
        input_shape: shape of the input dataset
    """
    global _models_built
    global _generator
    global _discriminator
    if not _initialized:
        raise RuntimeError(
            "Trainer not initialized. Please call `Trainer.init` method "
            "before calling `build`. "
        )

    batch_size, input_dim = input_shape
    total_cats = _metadata["categorical"]["total_num_categories"]
    # Generator receives the input of dimensions = data_dim + |cond_vector|
    # where, |cond_vector| = total_num_categories
    discriminator_input_shape = (batch_size, input_dim + total_cats)
    _discriminator.build(discriminator_input_shape)
    # Generator receives the input of dimensions = latent_dim + |cond_vector|
    # where, |cond_vector| = total_num_categories
    generator_input_shape = (batch_size, _latent_dim + total_cats)
    _generator.build(generator_input_shape)

    _models_built = True


def metrics():
    """
    Returns:
        List of metrics.
    """
    return [_generator_loss_tracker,
            _discriminator_loss_tracker]


def fit(x, epochs: int = 1, verbose: bool = True):
    """
    Trainer function for GAIN for JAX backend.

    Args:
        x: Dataset. An instance of `CTGANDataSampler` which can be accessed
            through `teras.preprocessing` package.
        epochs: int, Number of epochs to train.
        verbose: bool, whether to print training logs to console.

    Returns:
        A tuple of trained `generator_state` and `discriminator_state`.
    """
    global _models_built
    global _initialized
    global _compiled
    global _metadata
    global _latent_dim
    global _seed

    if not _initialized:
        raise RuntimeError(
            "Trainer not initialized. Please call `Trainer.init` method "
            "before calling `fit`. "
        )

    if not _compiled:
        raise RuntimeError(
            "Trainer not compiled. Please call `Trainer.compile()` method "
            "before calling `fit`. "
        )

    _optimizers_built = False
    total_batches = 0
    if dataset_type(x) == "not_supported":
        raise ValueError(
            "Unsupported type for `x`. "
            "It should be tensorflow dataset, pytorch dataloader."
            f"But received {type(x)}")
    aux_args = (_metadata, _latent_dim)
    key = jax.random.PRNGKey(_seed)
    is_first_iteration = True
    for epoch in range(epochs):
        epoch_start_time = time.time()
        for batch_num, batch in enumerate(x):
            if isinstance(batch, tuple):
                batch = (np.asarray(elem) for elem in batch)
            else:
                batch = np.asarray(batch)
            if is_first_iteration:
                if not _models_built:
                    if isinstance(batch, tuple):
                        input_shape = ops.shape(batch[0])
                    else:
                        input_shape = ops.shape(batch)
                    input_shape = (input_shape[:-1], input_shape[-1] * 2)
                    _generator.build(input_shape)
                    _discriminator.build(input_shape)
                    _models_built = True

                if not _optimizers_built:
                    _generator_optimizer.build(_generator.trainable_variables)
                    _discriminator_optimizer.build(
                        _discriminator.trainable_variables)
                    _optimizers_built = True

                # Create initial state
                generator_state = (
                    _generator.trainable_variables,
                    _generator.non_trainable_variables,
                    _generator_optimizer.variables
                )

                discriminator_state = (
                    _discriminator.trainable_variables,
                    _discriminator.non_trainable_variables,
                    _discriminator_optimizer.variables
                )

                is_first_iteration = False

            # print("\nSTATES:\n"
            #       f"\tGenerator: {generator_state}\n"
            #       f"\tDiscriminator: {discriminator_state}"
            #       )
            logs, generator_state, discriminator_state = train_step(
                generator_state,
                discriminator_state,
                batch,
                aux_args,
                jax.random.split(key, 1)[0]
            )
            if verbose:
                epoch_str = f"Epoch {epoch + 1}/{epochs}"
                elapsed_time_str = (f"Elapsed "
                                    f"{time.time() - epoch_start_time:.2f}s")
                if total_batches > 0 and epoch > 0:
                    batch_str = f"Batch {batch_num + 1}/{total_batches}"
                else:
                    batch_str = f"Batch {batch_num + 1}/?"
                    total_batches += 1
                logs_str = (f"generator_loss: "
                            f"{logs['generator_loss']:.4f}   "
                            f"discriminator_loss: "
                            f"{logs['discriminator_loss']:.4f}   "
                            )
                print(
                    f"\r{epoch_str:<15} {elapsed_time_str:<15} {batch_str:<15}"
                    f" {logs_str}",
                    end="")
        print()
        _trained = True

    return generator_state, discriminator_state


def generator_compute_loss_and_updates(
        generator_trainable_vars,
        generator_non_trainable_vars,
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        noise,
        cond_vectors,
        mask,
        metadata
):
    x_generated = _generator.stateless_call(
        generator_trainable_vars,
        generator_non_trainable_vars,
        noise)
    x_generated = ops.concatenate(
        [x_generated, cond_vectors], axis=1)
    y_pred_generated = _discriminator.stateless_call(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_generated)
    loss = ctgan_generator_loss(x_generated, y_pred_generated,
                                cond_vectors=cond_vectors, mask=mask,
                                metadata=metadata)
    return loss, (generator_non_trainable_vars,)


def discriminator_compute_grad_penalty_and_updates(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x,
        x_generated,
):
    y_pred_generated = _discriminator.stateless_call(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_generated)
    y_pred_real = _discriminator.stateless_call(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x)
    grad_pen = _discriminator.gradient_penalty(x, x_generated)
    return grad_pen, (discriminator_non_trainable_vars,)


def discriminator_compute_loss_and_updates(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x,
        x_generated,
):
    y_pred_generated = _discriminator.stateless_call(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x_generated)
    y_pred_real = _discriminator.stateless_call(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x)
    loss = ctgan_discriminator_loss(y_pred_real, y_pred_generated)
    return loss, (discriminator_non_trainable_vars,)


@jax.jit
def train_step(generator_state, discriminator_state, data, aux_args, key):
    (
        generator_trainable_vars,
        generator_non_trainable_vars,
        generator_optimizer_vars,
    ) = generator_state

    (
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        discriminator_optimizer_vars,
    ) = discriminator_state

    (
        metadata,
        latent_dim,
    ) = aux_args

    x, cond_vectors_real, cond_vectors, mask = data
    batch_size = ops.shape(x)[0]

    # =========================
    # Discriminator train step
    # =========================
    z = random.normal(shape=[batch_size, latent_dim],
                      seed=key)
    input_gen = ops.concatenate([z, cond_vectors], axis=1)
    x_generated, _ = _generator.stateless_call(
        generator_trainable_vars,
        generator_non_trainable_vars,
        input_gen)
    x_generated = ops.concatenate(
        [x_generated, cond_vectors], axis=1)
    x = ops.concatenate([x, cond_vectors_real],
                        axis=1)
    # w.r.t. gradient penalty
    disc_grad_fn_pen = jax.value_and_grad(
        discriminator_compute_grad_penalty_and_updates,
        has_aux=True)
    (grad_pen, (discriminator_non_trainable_vars,)), grads = disc_grad_fn_pen(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x,
        x_generated
    )
    (
        discriminator_trainable_vars,
        discriminator_optimizer_vars
    ) = _discriminator_optimizer.stateless_apply(
        discriminator_optimizer_vars,
        grads,
        discriminator_trainable_vars
    )
    # w.r.t. loss
    disc_grad_fn_loss = jax.value_and_grad(
        discriminator_compute_loss_and_updates,
        has_aux=True)
    (d_loss, (discriminator_non_trainable_vars,)), grads = disc_grad_fn_loss(
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        x,
        x_generated
    )
    (
        discriminator_trainable_vars,
        discriminator_optimizer_vars
    ) = _discriminator_optimizer.stateless_apply(
        discriminator_optimizer_vars,
        grads,
        discriminator_trainable_vars
    )

    discriminator_state = (
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        discriminator_optimizer_vars
    )

    # =====================
    # Generator train step
    # =====================
    z = random.normal(shape=[batch_size, latent_dim],
                      seed=jax.random.split(key, 1)[0])
    input_gen = ops.concatenate([z, cond_vectors], axis=1)
    gen_grad_fn = jax.value_and_grad(
        generator_compute_loss_and_updates,
        has_aux=True,
    )
    (g_loss, (generator_non_trainable_vars,)), grads = gen_grad_fn(
        generator_trainable_vars,
        generator_non_trainable_vars,
        discriminator_trainable_vars,
        discriminator_non_trainable_vars,
        input_gen,
        cond_vectors,
        mask,
        metadata
    )

    (
        generator_trainable_vars,
        generator_optimizer_vars
    ) = _generator_optimizer.stateless_apply(
        generator_optimizer_vars,
        grads,
        generator_trainable_vars
    )

    generator_state = (
        generator_trainable_vars,
        generator_non_trainable_vars,
        generator_optimizer_vars
    )

    # Update custom tracking metrics
    _generator_loss_tracker.update_state(g_loss)
    _discriminator_loss_tracker.update_state(d_loss)

    logs = {m.name: m.result() for m in metrics()}
    return logs, generator_state, discriminator_state
